// $Id: DAMLDatatypeValidator.java,v 1.4 2003/03/11 22:55:58 drager Exp $

package org.daml.validator;
import java.util.HashSet;
import java.util.HashMap;
import java.util.Iterator;
import org.daml.validator.xsd.DatatypeValidator;
import org.daml.validator.xsd.imp.DatatypeValidatorImp;
import org.daml.validator.xsd.UnknownSchemaException;
import org.daml.validator.xsd.SchemaValidationException;
import org.daml.validator.xsd.UnknownSimpleTypeException;

public class DAMLDatatypeValidator {

    private boolean DEBUG = false;
    private boolean ERRORS = true;
    
    /**
     * Namespaces cached in the Datatype Validator.
     */
    HashSet datatypeURIs = new HashSet();

    /**
     * XML Schema Datatype Validator
     */
    DatatypeValidator datatypeValidator;

    /**
     * Mapping from URIs to the corresponding Cache URI.
     * The XML Schema Datatype Validator stores the datatypes under the
     * namespace of the URI that was loaded.  If it came from the cache
     * then we need to do the lookup from the cache.  This stores
     * the mapping for use during lookup.
     */
    private HashMap cacheMap = new HashMap();

    /**
     * Creates an XML Schema Datatype Validator for the DAML Validator.
     * Relies heavily of the XML Schema Datatype Validator by Mark Neighbors.
     * XML Schema Datatype files are added to the validator.  Once added 
     * datatypes can be validated against the datatype definitions.
     */
    public DAMLDatatypeValidator() 
    {
	this.initialize();
    }

    /**
     * Initialze or reset the DAML Datatype Validator.
     * Removes all datatype definitions.
     */
    public void initialize () 
    {
	datatypeValidator = new DatatypeValidatorImp();
	datatypeURIs.clear();
	datatypeURIs.addAll(datatypeValidator.getXMLSchemaSchemaURLSet());
	cacheMap.clear();
    }

    /**
     * Returns true if the URL is an XML Schema file.
     */
    public boolean isASchema (String url)
    {
	boolean result = false;
	boolean ERRORS = false;
	try {
	    result = datatypeValidator.isASchema(url);
	} catch (UnknownSchemaException e1) {
	    if (ERRORS) System.out.println("isASchema("+url+") -> "+e1);
	} catch (SchemaValidationException e2) {
	    if (ERRORS) System.out.println("isASchema("+url+") -> "+e2);
	}
	if (DEBUG) System.out.println("isASchema("+url+") -> "+result);
	return result;
    }

    /**
     * Add a XML Schema Datatype definition file to the validator.
     */
    public void add (String uri) 
	throws UnknownSchemaException, SchemaValidationException
    {
	if (DEBUG) System.out.println("add("+uri+")");
	datatypeValidator.assertValidSchema(stripHash(uri));
	datatypeURIs.add(stripHash(uri));
    }

    /**
     * Add a XML Schema Datatype definition file to the validator.
     */
    public void addFromCache (String uri, String cacheURI) 
	throws UnknownSchemaException, SchemaValidationException
    {
	if (DEBUG) System.out.println("addFromCache("+uri+", "+cacheURI+")");
	datatypeValidator.assertValidSchema(stripHash(cacheURI));
	datatypeURIs.add(stripHash(uri));
	cacheMap.put(stripHash(uri), stripHash(cacheURI));
    }

    /**
     * Returns true if the namespace URI is a XML Schema Datatype namespace
     * that has already been added (or is a default XML Schema Schema URL).
     */
    public boolean hasNamespace (String uri)
    {
	boolean result = datatypeURIs.contains(stripHash(uri));
	if (DEBUG) System.out.println("hasNamespace("+uri+") -> "+result);
	return result;
    }

    /**
     * Returns true if a datatype definition exists for the given datatype.
     */
    public boolean isAKnownSimpleType (String namespace, String datatype)
    {
	boolean result = false;
	String ns = stripHash(namespace);
	try {
	    if (hasNamespace(ns)) {
		result = datatypeValidator.isAKnownSimpleType(checkNamespace(ns), datatype);
	    }
	} catch (UnknownSchemaException e1) {
	    if (ERRORS) System.out.println("isAKnownSimpleType("+namespace+", "+datatype+") -> "+e1);
	} catch (SchemaValidationException e2) {
	    if (ERRORS) System.out.println("isAKnownSimpleType("+namespace+", "+datatype+") -> "+e2);
	}
	if (DEBUG) System.out.println("isAKnownSimpleType("+namespace+", "+datatype+") -> "+result);
	return result;
    }
	
    /**
     * Returns true if the value is a proper instance of the given datatype.
     */
    public boolean isAValidValue( String value, String namespace, String datatype)
	throws UnknownSimpleTypeException  
    {
	boolean result = false;
	String ns = stripHash(namespace);
	try {
	    if (hasNamespace(ns)) {
		result = datatypeValidator.isAValidValue(value, checkNamespace(ns), datatype);
	    } else {
		throw(new UnknownSimpleTypeException(ns+"#"+datatype));
	    }
	} catch (UnknownSchemaException e1) {
	    if (ERRORS) System.out.println("isAValidValue("+value+", "+namespace+", "+datatype+") -> "+e1);
	} catch (SchemaValidationException e2) {
	    if (ERRORS) System.out.println("isAValidValue("+value+", "+namespace+", "+datatype+") -> "+e2);
	}
	if (DEBUG) System.out.println("isAValidValue("+value+", "+namespace+", "+datatype+") -> "+result);
	return result;

    }

    /**
     * Returns true if the value and type pair is a proper instance if the given datatype.
     */
    public boolean isAValidTypedValue(String valueNamespace, String valueDatatype, String value, String namespace, String datatype)
	throws UnknownSimpleTypeException  
    {
	boolean result = false;
	String ns = stripHash(namespace);
	try {
	    if (hasNamespace(ns)) {
		String vns = checkNamespace(stripHash(valueNamespace));
		result = datatypeValidator.isAValidTypedValue(vns, valueDatatype, value, checkNamespace(ns), datatype);
	    } else {
		throw(new UnknownSimpleTypeException(ns+"#"+datatype));
	    }
	} catch (UnknownSchemaException e1) {
	    if (ERRORS) System.out.println("isAValidTypedValue("+valueNamespace+", "+valueDatatype+", "+value+", "+namespace+", "+datatype+") -> "+e1);
	} catch (SchemaValidationException e2) {
	    if (ERRORS) System.out.println("isAValidTypedValue("+valueNamespace+", "+valueDatatype+", "+value+", "+namespace+", "+datatype+") -> "+e2);
	}
	if (DEBUG) System.out.println(">> isAValidTypedValue("+valueNamespace+", "+valueDatatype+", "+value+", "+namespace+", "+datatype+") -> "+result);
	return result;
    }


    private String checkNamespace(String namespace)
    {
	String ns = stripHash(namespace);
	if (cacheMap.containsKey(ns)) {
	    return (String)cacheMap.get(ns);
	} else {
	    return ns;
	}
    }

    /**
     * Returns the namespace string with any trailing '#' character removed.
     */
    private String stripHash(String namespace)
    {
	if (namespace == null) return null;
	int length = namespace.length();
	if (namespace.charAt(length-1) == '#') {
	    return namespace.substring(0, length-1);
	} else {
	    return namespace;
	}
    }


    public HashSet getXMLSchemaSchemaURLSet()
    {
	return datatypeValidator.getXMLSchemaSchemaURLSet();
    }

    public HashSet getSimpleTypeSet()
    {
	boolean DEBUG = false;
	HashSet result = new HashSet();
	HashSet datatypes = ((DatatypeValidatorImp)datatypeValidator).getSimpleTypeSet();
	for (Iterator i = datatypes.iterator(); i.hasNext(); ) {
	    String dt = (String)i.next();
	    if (DEBUG) System.out.print("-> "+dt+" ");
	    int hashIndex = dt.lastIndexOf('#');
	    if (hashIndex >= 0) {
		String ns = dt.substring(0, hashIndex);
		if (cacheMap.containsValue(ns)) {
		    // Find the original namespace for the cached namespace
		    boolean found = false;
		    for (Iterator j = cacheMap.entrySet().iterator(); j.hasNext() && !found; ) {
			java.util.Map.Entry entry = (java.util.Map.Entry)j.next();
			if (ns.equals(entry.getValue())) {
			    found = true;
			    String tmp = entry.getKey() + dt.substring(hashIndex);
			    result.add(tmp);
			    if (DEBUG) System.out.println(tmp);
			}
		    }
		} else {
		    // not in the cache
		    result.add(dt);
		    if (DEBUG) System.out.println(".");
		}
	    } else {
		// Unexpected: no # in datatype URI
		result.add(dt);
		if (DEBUG) System.out.println(".");
	    }		
	}
	return result;
    }
}

