// $Id: Indication.java,v 1.17 2003/03/11 22:55:58 drager Exp $

package org.daml.validator.indications;

//import org.w3c.rdf.model.*;

import java.util.*;
import org.daml.validator.model.*;
import com.hp.hpl.mesa.rdf.jena.model.Resource;
import com.hp.hpl.mesa.rdf.jena.model.Statement;
import com.hp.hpl.mesa.rdf.jena.model.RDFException;
import com.hp.hpl.mesa.rdf.jena.common.StatementImpl;
import com.hp.hpl.jena.rdf.arp.Location;


/** 
 * Basic validator Indication object used to specify problems found by the validator.
 * Specific indications are subclasses.
 * A basic indication contains the indication type, a level, and a message string.
 * There are three indication levels: Indication.ERROR, Indication.WARNING,
 * and Indication.INFORMATION.  Each indication is assigned a unique id number.
 */
public class Indication
{
    /** Error level type */
    public static final int ERROR = 0;
    /** Warning level type */
    public static final int WARNING = 1;
    /** Information level type */
    public static final int INFORMATION = 2;

    /**
     * The unique id for an indication
     */
    int id;

    /**
     * The id counter used for assigning unique ids to indications
     */
    private static int idCounter = 0;

    /**
     * The number of individual indications.
     * Occurances of compound indications are counted separately.
     */
    protected static int totalIndications = 0;

    /**
     * The readable name for this indication.
     */
    String type = "General Indication";

    /**
     * The default level for this indication class.
     */
    int defaultLevel = ERROR;

    // Basic values for an instantiated Indication.

    /**
     * The level of the indication. 
     */
    int level = ERROR;

    /**
     * String describing the indication.
     */
    String message = "No additional information specified.";

    /**
     * Generic indication constructor without message.
     */
    public Indication()
    {
	level = defaultLevel;
	id = ++idCounter;
	totalIndications++;
    }

    /**
     * Generic indication constructor with message.
     * @param message	A description string for the indication.
     */
    public Indication(String message)
    {
	this.message = message;
	level = defaultLevel;
	id = ++idCounter;
	totalIndications++;
    }

    /**
     * Generate a message describing the indication.
     * This method should be customized by subClasses.
     * @return A message describing the indication.
     */
    public String getMessage()
    {
	if (message == null) {
	    return("");
	} else {
	    return(message);
	}
    }

    /**
     * Generate a location string for the indication.
     * @return Location string
     */
    public String getLocation()
    {
	return("");
    }

    /**
     * Generate a string representation of the indication.
     * @return The string representation.
     */
    public String toString()
    {
	String levelStr = null;
	if (level == ERROR) {
	    levelStr = "ERROR - ";
	} else if (level == WARNING) {
	    levelStr = "WARNING - ";
	} else if (level == INFORMATION) {
	    levelStr = "INFORMATION - ";
	}

	// Construct the indication message string
	String m = "[" + id + "] " + levelStr + type +": ";

	m += getMessage() + "\n";
	
	m += getLocation();

	return(m);
    }

    /**
     * Generate a string that can be used to compare indications.
     * The ids are excluded because the may not be consistant 
     * between runs of the validator.
     * @return The comparable string representation.
     */
    public String toComparableString()
    {
	String levelStr = null;
	if (level == ERROR) {
	    levelStr = "ERROR - ";
	} else if (level == WARNING) {
	    levelStr = "WARNING - ";
	} else if (level == INFORMATION) {
	    levelStr = "INFORMATION - ";
	}

	// Construct the indication message string
	String m = levelStr + type +": ";

	m += getMessage() + "\n";
	
	m += getLocation();

	return(m);
    }


    /**
     * Set the level of the indication
     * @param level An Indication level value.
     */
    public void setLevel(int level) 
    {
	this.level = level;
    }

    /**
     * Generate the XML representation of the indication.  
     * @returns The indication as XML.
     */ 
    public String toXML()
    throws RDFException
    {
	String levelStr = "error";
	if (level == ERROR) {
	    levelStr = "error";
	} else if (level == WARNING) {
	    levelStr = "warning";
	} else if (level == INFORMATION) {
	    levelStr = "information";
	}

	levelStr =  "<level>" + levelStr + "</level>\n";

	String typeStr = "<type>"+type+"</type>\n";

	String locationStr = this.getLocationXML();

	String idStr = "<id>" + id + "</id>\n";

	String messageStr = "<message>" + fixXMLString(getMessage()) + "</message>\n";

	String xml = "<indication>\n";
	xml += idStr + typeStr + levelStr + locationStr + messageStr;
	xml +="</indication>\n";
	
	return(xml);
    }

    /**
     * Convert reserved XML characeters
     * < becomes &lt;
     * & becomes &amp;
     * > becomes &gt;
     * ' becomes &apos;
     * " becomes &quot;
     * @param The string to convert.
     * @return The string with reserved XML characters converted.
     */
    // A simple function, but which implementation is more efficient?
    private static String fixXMLString(String s) 
    {        
	s = s.replaceAll("&", "&amp;");
	s = s.replaceAll("<", "&lt;");
	s = s.replaceAll(">", "&gt;");
	s = s.replaceAll("'", "&apos;");
	s = s.replaceAll("\t", "&#9;");
	s = s.replaceAll("\n", "<br/>");
	s = s.replaceAll("\r", "<br/>");
	s = s.replaceAll("\"", "&quot;");
	return s;
    }

    /**
     * Reset the counter used to assign indication ids to 0.
     */
    public static void resetIdCounter()
    {
	idCounter = 0;
	totalIndications = 0;
    }

    /**
     * Return the number of indications created since the id counter
     * was last reset.
     * @return The number of indications.
     */
    public static int indicationCount()
    {
	return(totalIndications);
    }

    /**
     * Generate XML for the location of an exception without any location.
     * @return Location as XML
     */
    public String getLocationXML()
	throws RDFException
    {
	return("");
    }
     
    
    /**
     * Generate a location XML string from a collection of statements.
     * Organizes the statements by URI.
     * @param statements A collection of RDF statements.
     * @return The statment locations as XML.
     */
    public String getLocationXML(Collection statements)
	throws RDFException
    {		
	String xml = "";
	// Organize statements by URI
	HashMap ht = new HashMap(statements.size());
	for (Iterator i = statements.iterator(); i.hasNext(); ) {
	    Object o = i.next();
	    if (o instanceof Statement) {
		// Get the URI 
		String uri = "Unknown URI";
// 		if (o instanceof StatementImpl) {
// 		    uri = ((StatementImpl)o).getSourceURI();
// 		}
		// Lookup set of statements for URI
		HashSet h = (HashSet)ht.get(uri);
		// Create a set if needed
		if (h == null) {
		    h = new HashSet();
		    ht.put(uri, h);
		}
		// Add statement to the set.
		h.add(o);
	    } else {
		// For some reason this is not a statement
		System.out.println("DEBUG: Didn't find a statement: "+o);
	    }
	}   
	// Generate XML, grouping statements by URI
	for (Iterator e = ht.keySet().iterator(); e.hasNext(); ) {
	    String uri = (String)e.next();
	    HashSet h = (HashSet)ht.get(uri);
	    xml += "<location>\n";
	    xml += "  <uri>"+uri+"</uri>\n";
	    for (Iterator i = h.iterator(); i.hasNext(); ) {
		Statement s = (Statement)i.next();
		xml += getStatementXML(s);
	    } 
	    xml += "</location>\n";
	}
	return(xml);
    }


    /**
     * Generate the location XML string for a node.
     * @param node A DAML node
     * @return The XML location for the node.
     */
    public String getLocationXML(DAMLModelImpl.Node node)
    throws RDFException
    {
	String xml = "";
	Statement stmt = node.getTypeStatement();
// 	if (stmt instanceof StatementImpl) {
// 	    StatementImpl s = (edu.stanford.db.rdf.model.i.StatementImpl)stmt;
// 	    xml += "<location>\n";
// 	    xml += "<resource>\n";
// 	    xml += "<uri>"+stmt.getSubject()+"</uri>\n";
// 	    xml += "<line>"+s.getLineNumber()+"</line>\n";
// 	    xml += "<column>"+s.getColumnNumber()+"</column>\n";
// 	    xml += "</resource>\n";
// 	    xml += "</location>\n";
// 	}
	return(xml);
    }

    /**
     * Generate a location xml string for a statement.
     * @param s An RDF statement.
     * @return The XML location for the statement.
     */
    public String getLocationXML(Statement s)
	throws RDFException
    {
	String uri = "Unknown URI";
// 	if (s instanceof StatementImpl) {
// 	    try {
// 		uri = ((StatementImpl)s).getSourceURI();
// 	    } catch (java.lang.NoSuchMethodError ex) {
// 	    }
// 	}
	String xml = "<location>\n";
	xml += "  <uri>"+uri+"</uri>\n";
	xml += getStatementXML(s);
	xml += "</location>\n";
	return(xml);
    }

    /**
     * Generate a location xml string for a URI.
     * @param uri URI string.
     * @return The XML location for the URI.      
     */
    public String getLocationXML(String uri)
	throws RDFException
    {
	String xml ="<location>\n";
	xml += "<resource>\n";
	xml += "  <uri>"+uri+"</uri>";
	xml += "</resource>\n";
	xml += "</location>\n";
	return(xml);
    }

    /**
     * Generate a statement xml string for Statement.
     * @param s an RDF Statement
     * @return The XML representation of the statement.
     */
    public String getStatementXML(Statement s)
	throws RDFException
    {
	String xml = "<statement>\n";
// 	if (s instanceof StatementImpl) {
// 	    xml += "  <line>"+((StatementImpl)s).getLineNumber()+"</line>\n";
// 	    xml += "  <column>"+((StatementImpl)s).getColumnNumber()+"</column>\n";
// 	}
	xml += "  <predicate>" + s.getPredicate() + "</predicate>\n";
	xml += "  <subject>" + s.getSubject() + "</subject>\n";
	if (s.getObject() instanceof Resource) {
	    xml += "  <objectresource>" + s.getObject() + "</objectresource>\n";
	} else {
	    xml += "  <objectliteral>" + fixXMLString(s.getObject().toString()) + "</objectliteral>\n";
	}	    
	xml += "</statement>\n";
	return(xml);
    }

    /**
     * Returns a location string for a URI.
     * @param uri A URI string.
     * @return The location string for the URI.
     */
    public String getLocation(String uri)
    {
	return("In "+uri);
    }	

    /**
     * Returns a location string for a statement.
     * @param s An RDF statement.
     * @return The location strint for the statement.
     */
    public String getLocation(Statement s) 
    {
	String location = "";
	location += "At "+s;
	Location l = s.getLocation();

	if (l != null) {
	    location += " "+l.toString();
	}
	location += "\n";

	return(location);
    }

    /**
     * Returns a location string for a collection of statements.
     * The statements are grouped by URI and the returned list is
     * sorted by statement and URI.
     * @param statements A collection of RDF statements.
     * @return The string representation for the location of the statements.
     */
    public String getLocation(Collection statements) 
    {
	String location = "";
	// Organize statements by URI
	TreeMap ht = new TreeMap();
	for (Iterator i = statements.iterator(); i.hasNext(); ) {
	    Object o = i.next();
	    if (o instanceof Statement) {
		// Get the URI 
		String uri = "Unknown URI";
// 		if (o instanceof StatementImpl) {
// 		    try {
// 			uri = ((StatementImpl)o).getSourceURI();
// 		    } catch (java.lang.NoSuchMethodError ex) {
// 		    }
// 		}
		// Lookup set of statements for URI
		TreeSet h = (TreeSet)ht.get(uri);
		// Create a set if needed
		if (h == null) {
		    h = new TreeSet();
		    ht.put(uri, h);
		}
		// Add statement string to the set.
		h.add(o.toString());
	    } else {
		// For some reason this is not a statement
		System.out.println("DEBUG: Didn't find a statement: "+o);
	    }
	}
	// Generate string, grouping statements by URI
	for (Iterator e = ht.keySet().iterator(); e.hasNext(); ) {
	    String uri = (String)e.next();
	    TreeSet h = (TreeSet)ht.get(uri);
	    location +="In "+uri+"\n";
	    for (Iterator i = h.iterator(); i.hasNext(); ) {
		String s = (String)i.next();
		location += " at "+s+"\n";
	    }
	}
	return(location);
    }

}
